/*-------------------------------------------------------------------------
 Asynchronous FIFO, 8-bit, 15 depth
 
 File name   : syncfifo_8x15.v
 Version     : Version 1.1
 Created     : Jun/19/2005
 Last update : Sep/01/2009
 Desgined by : Toshihiro Katashita
 
 
 Copyright (C) 2005, 2009 AIST
 
 By using this code, you agree to the following terms and conditions.
 
 This code is copyrighted by AIST ("us").
 
 Permission is hereby granted to copy, reproduce, redistribute or
 otherwise use this code as long as: there is no monetary profit gained
 specifically from the use or reproduction of this code, it is not sold,
 rented, traded or otherwise marketed, and this copyright notice is
 included prominently in any copy made.
 
 We shall not be liable for any damages, including without limitation
 direct, indirect, incidental, special or consequential damages arising
 from the use of this code.
 
 When you publish any results arising from the use of this code, we will
 appreciate it if you can cite our webpage
 (http://www.rcis.aist.go.jp/special/SASEBO/).
 -------------------------------------------------------------------------*/ 


//================================================ ASYNCFIFO_8x15
module ASYNCFIFO_8x15
  (wd, we, ful, aful, wclk, wrst,
   rd, re, emp, aemp, rclk, rrst);

   //------------------------------------------------
   parameter WIDTH = 8;
   parameter DEPTH = 5;
   parameter MEM_DEPTH = (1<<DEPTH)-1;
   input [WIDTH-1:0]  wd;         // Write data
   input 	      we;         // Write enable
   output 	      ful, aful;  // FIFO full / almost full
   input 	      wclk, wrst; // Clock and reset for write port
   output [WIDTH-1:0] rd;         // Read data
   input 	      re;         // Read enable
   output 	      emp, aemp;  // FIFO empty / almost empty
   input 	      rclk, rrst; // Clock and reset for read port
   
   //------------------------------------------------
   // write
   wire 	      update_ful, update_aful;
   wire 	      async_ful,  async_aful, async_bful;
   reg 		      ful,        aful;
   wire 	      wr_enable;
   reg [DEPTH-1:0]    wa, wa_gray1, wa_gray2, wa_gray3;

   // read
   wire 	      update_emp, update_aemp;
   wire 	      async_emp,  async_aemp, async_bemp;
   reg 		      emp,        aemp;
   wire 	      rd_enable;
   reg [DEPTH-1:0]    ra, ra_gray1, ra_gray2, ra_gray3, ra_gray4;

   //------------------------------------------------
   //synthesis attribute ram_style of mem is distributed;
   reg [WIDTH-1:0]    mem [0:MEM_DEPTH];
   reg [DEPTH-1:0]    ra1;

   //------------------------------------------------
   // FIFO write
   assign update_ful  = (we | ful);
   assign update_aful = (we | aful);
   assign wr_enable   = (we & ~ful);

   // ful flag(registered), aful flag(registered)
   always @(posedge wclk or posedge wrst)
     if (wrst)            ful <= 1;
     else if (update_ful)
       if (ful)           ful <= async_ful;
       else               ful <= async_aful;

   always @(posedge wclk or posedge wrst)
     if (wrst)             aful <= 1;
     else if (update_aful)
       if (ful)            aful <= 1;
       else if (aful)      aful <= async_aful | we;
       else                aful <= async_bful;
   
   // memory write address
   always @(posedge wclk or posedge wrst)
     if (wrst)           wa <= 0;
     else if (wr_enable) wa <= wa + 1;

   // memory write address (gray code) (-1, -2, -3)
   always @(posedge wclk or posedge wrst)
     if (wrst)           wa_gray1 <= bin2gray(MEM_DEPTH);
     else if (wr_enable) wa_gray1 <= bin2gray(wa);
   always @(posedge wclk or posedge wrst)
     if (wrst)           wa_gray2 <= bin2gray(MEM_DEPTH-1);
     else if (wr_enable) wa_gray2 <= wa_gray1;
   always @(posedge wclk or posedge wrst)
     if (wrst)           wa_gray3 <= bin2gray(MEM_DEPTH-2);
     else if (wr_enable) wa_gray3 <= wa_gray2;

   //------------------------------------------------
   // FIFO read
   assign update_emp  = re | emp;
   assign update_aemp = re | aemp;
   assign rd_enable   = re & ~emp;
   
   // emp flag(registered), aemp flag(registered)
   always @(posedge rclk or posedge rrst)
     if (rrst)            emp <= 1;
     else if (update_emp)
       if (emp)           emp <= async_emp;
       else               emp <= async_aemp;
   
   always @(posedge rclk or posedge rrst)
     if (rrst)            aemp <= 1;
     else if (update_aemp)
       if (emp)           aemp <= 1;
       else if (aemp)     aemp <= async_aemp | re;
       else               aemp <= async_bemp;
   
   // memory read address
   always @(posedge rclk or posedge rrst)
     if (rrst)           ra <= 0;
     else if (rd_enable) ra <= ra + 1;

   // memory read address (gray code) (-1, -2, -3, -4)
   always @(posedge rclk or posedge rrst)
     if (rrst)           ra_gray1 <= bin2gray(MEM_DEPTH);
     else if (rd_enable) ra_gray1 <= bin2gray(ra);
   always @(posedge rclk or posedge rrst)
     if (rrst)           ra_gray2 <= bin2gray(MEM_DEPTH-1);
     else if (rd_enable) ra_gray2 <= ra_gray1;
   always @(posedge rclk or posedge rrst)
     if (rrst)           ra_gray3 <= bin2gray(MEM_DEPTH-2);
     else if (rd_enable) ra_gray3 <= ra_gray2;
   always @(posedge rclk or posedge rrst)
     if (rrst)           ra_gray4 <= bin2gray(MEM_DEPTH-3);
     else if (rd_enable) ra_gray4 <= ra_gray3;

   //------------------------------------------------
   // bynary code to gray code
   function [DEPTH-1:0] bin2gray;
      input [DEPTH-1:0] bin;
      integer i;
      begin
	 for (i=0;i<(DEPTH-1);i=i+1)
	   bin2gray[i] = bin[i+1] ^ bin[i];
	 bin2gray[DEPTH-1] = bin[DEPTH-1];
      end
   endfunction // bin2gray

   //------------------------------------------------
   // asynchronous ful flag
   assign async_ful  = (wa_gray3 == ra_gray4);
   assign async_aful = (wa_gray2 == ra_gray4);
   assign async_bful = (wa_gray1 == ra_gray4);

   // asynchronous emp flag
   assign async_emp  = (ra_gray3 == wa_gray3);
   assign async_aemp = (ra_gray2 == wa_gray3);
   assign async_bemp = (ra_gray1 == wa_gray3);

   //------------------------------------------------
   // FIFO memory
   always @(posedge wclk) if (wr_enable) mem[wa] <= wd;
   always @(posedge rclk) ra1 <= ra;
   assign rd = mem[ra1];
endmodule // ASYNCFIFO_8x15
